package java.util.concurrent;

import java.util.NoSuchElementException;
import java.util.Iterator;
import java.util.Collection;
import java.util.concurrent.locks.Condition;
import java.util.PriorityQueue;
import java.util.concurrent.locks.ReentrantLock;
import java.util.AbstractQueue;

public class DelayQueue<E extends Delayed> extends AbstractQueue<E> implements BlockingQueue<E>
{
    private final transient ReentrantLock lock;
    private final PriorityQueue<E> q;
    private Thread leader;
    private final Condition available;
    
    public DelayQueue() {
        this.lock = new ReentrantLock();
        this.q = new PriorityQueue();
        this.leader = null;
        this.available = this.lock.newCondition();
    }
    
    public DelayQueue(Collection<? extends E> c) {
        this.lock = new ReentrantLock();
        this.q = new PriorityQueue();
        this.leader = null;
        this.available = this.lock.newCondition();
        this.addAll(c);
    }
    
    @Override
    public boolean add(E e) {
        return this.offer(e);
    }
    
    @Override
    public boolean offer(E e) {
        ReentrantLock lock = this.lock;
        lock.lock();
        try {
            this.q.offer(e);
            if (this.q.peek() == e) {
                this.leader = null;
                this.available.signal();
            }
            return true;
        }
        finally {
            lock.unlock();
        }
    }
    
    @Override
    public void put(E e) {
        this.offer(e);
    }
    
    @Override
    public boolean offer(E e, long n, TimeUnit timeUnit) {
        return this.offer(e);
    }
    
    @Override
    public E poll() {
        ReentrantLock lock = this.lock;
        lock.lock();
        try {
            Delayed delayed = (E)this.q.peek();
            if (delayed == null || delayed.getDelay(TimeUnit.NANOSECONDS) > 0L) {
                return null;
            }
            return (E)this.q.poll();
        }
        finally {
            lock.unlock();
        }
    }
    
    @Override
    public E take() throws InterruptedException {
        ReentrantLock lock = this.lock;
        lock.lockInterruptibly();
        while (true) {
            try {
                Delayed delayed = (E)this.q.peek();
                Block_3: {
                    if (delayed == null) {
                        this.available.await();
                    }
                    else {
                        long delay = delayed.getDelay(TimeUnit.NANOSECONDS);
                        if (delay <= 0L) {
                            break Block_3;
                        }
                        if (this.leader != null) {
                            this.available.await();
                        }
                        else {
                            Thread currentThread = Thread.currentThread();
                            this.leader = currentThread;
                            try {
                                this.available.awaitNanos(delay);
                            }
                            finally {
                                if (this.leader == currentThread) {
                                    this.leader = null;
                                }
                            }
                        }
                    }
                    continue;
                }
                return (E)this.q.poll();
            }
            finally {
                if (this.leader == null && this.q.peek() != null) {
                    this.available.signal();
                }
                lock.unlock();
            }
        }
    }
    
    @Override
    public E poll(long n, TimeUnit timeUnit) throws InterruptedException {
        long n2 = timeUnit.toNanos(n);
        ReentrantLock lock = this.lock;
        lock.lockInterruptibly();
        while (true) {
            try {
                Delayed delayed = (E)this.q.peek();
                if (delayed == null) {
                    if (n2 <= 0L) {
                        return null;
                    }
                    n2 = this.available.awaitNanos(n2);
                }
                else {
                    long delay = delayed.getDelay(TimeUnit.NANOSECONDS);
                    if (delay <= 0L) {
                        return (E)this.q.poll();
                    }
                    if (n2 <= 0L) {
                        return null;
                    }
                    if (n2 < delay || this.leader != null) {
                        n2 = this.available.awaitNanos(n2);
                    }
                    else {
                        Thread currentThread = Thread.currentThread();
                        this.leader = currentThread;
                        try {
                            n2 -= delay - this.available.awaitNanos(delay);
                        }
                        finally {
                            if (this.leader == currentThread) {
                                this.leader = null;
                            }
                        }
                    }
                }
                continue;
            }
            finally {
                if (this.leader == null && this.q.peek() != null) {
                    this.available.signal();
                }
                lock.unlock();
            }
        }
    }
    
    @Override
    public E peek() {
        ReentrantLock lock = this.lock;
        lock.lock();
        try {
            return (E)this.q.peek();
        }
        finally {
            lock.unlock();
        }
    }
    
    @Override
    public int size() {
        ReentrantLock lock = this.lock;
        lock.lock();
        try {
            return this.q.size();
        }
        finally {
            lock.unlock();
        }
    }
    
    private E peekExpired() {
        Delayed delayed = (E)this.q.peek();
        return (delayed == null || delayed.getDelay(TimeUnit.NANOSECONDS) > 0L) ? null : delayed;
    }
    
    @Override
    public int drainTo(Collection<? super E> collection) {
        if (collection == null) {
            throw new NullPointerException();
        }
        if (collection == this) {
            throw new IllegalArgumentException();
        }
        ReentrantLock lock = this.lock;
        lock.lock();
        try {
            int n;
            Delayed peekExpired;
            for (n = 0; (peekExpired = this.peekExpired()) != null; ++n) {
                collection.add(peekExpired);
                this.q.poll();
            }
            return n;
        }
        finally {
            lock.unlock();
        }
    }
    
    @Override
    public int drainTo(Collection<? super E> collection, int n) {
        if (collection == null) {
            throw new NullPointerException();
        }
        if (collection == this) {
            throw new IllegalArgumentException();
        }
        if (n <= 0) {
            return 0;
        }
        ReentrantLock lock = this.lock;
        lock.lock();
        try {
            int n2;
            Delayed peekExpired;
            for (n2 = 0; n2 < n && (peekExpired = this.peekExpired()) != null; ++n2) {
                collection.add(peekExpired);
                this.q.poll();
            }
            return n2;
        }
        finally {
            lock.unlock();
        }
    }
    
    @Override
    public void clear() {
        ReentrantLock lock = this.lock;
        lock.lock();
        try {
            this.q.clear();
        }
        finally {
            lock.unlock();
        }
    }
    
    @Override
    public int remainingCapacity() {
        return Integer.MAX_VALUE;
    }
    
    @Override
    public Object[] toArray() {
        ReentrantLock lock = this.lock;
        lock.lock();
        try {
            return this.q.toArray();
        }
        finally {
            lock.unlock();
        }
    }
    
    @Override
    public <T> T[] toArray(T[] a) {
        ReentrantLock lock = this.lock;
        lock.lock();
        try {
            return this.q.toArray(a);
        }
        finally {
            lock.unlock();
        }
    }
    
    @Override
    public boolean remove(Object o) {
        ReentrantLock lock = this.lock;
        lock.lock();
        try {
            return this.q.remove(o);
        }
        finally {
            lock.unlock();
        }
    }
    
    void removeEQ(Object o) {
        ReentrantLock lock = this.lock;
        lock.lock();
        try {
            Iterator<E> iterator = this.q.iterator();
            while (iterator.hasNext()) {
                if (o != iterator.next()) {
                    continue;
                }
                iterator.remove();
                break;
            }
        }
        finally {
            lock.unlock();
        }
    }
    
    @Override
    public Iterator<E> iterator() {
        return new DelayQueue.Itr(this.toArray());
    }
    
    private class Itr implements Iterator<E>
    {
        final Object[] array;
        int cursor;
        int lastRet;
        
        Itr(Object[] array) {
            this.lastRet = -1;
            this.array = array;
        }
        
        @Override
        public boolean hasNext() {
            return this.cursor < this.array.length;
        }
        
        @Override
        public E next() {
            if (this.cursor >= this.array.length) {
                throw new NoSuchElementException();
            }
            this.lastRet = this.cursor;
            return (Delayed)this.array[this.cursor++];
        }
        
        @Override
        public void remove() {
            if (this.lastRet < 0) {
                throw new IllegalStateException();
            }
            DelayQueue.this.removeEQ(this.array[this.lastRet]);
            this.lastRet = -1;
            return;
        }
    }
}
